/*****************************************************************
Counter Firmware for Geiger Mueller counter. Timer0 and tmr0_msb
variable are used to count events during a one second interval,
which are stored in counter. The number of counts per second are
stored in an array, later summed and multiplied by 6 to get the CPM.

Data is sent over the serial port using a master/slave protocol.
No data is sent until requested by the master. Count values are
sent as soon as they are ready, but only once cleared to do so by
the master. Specifications of the GM tube are hard-coded into the
firmware.

Update interval is 1000ms.

Currently used with a "SBM-19" Geiger tube. Conversion rate and
highest count rate are taken from:

http://www.pocketmagic.net/tube-sbm-19-%D1%81%D0%B1%D0%BC-19/
https://sites.google.com/site/diygeigercounter/gm-tubes-supported

External piezo buzzer is used to indicate when an event is detected.
Buzzer frequency is 4kHz. Buzzer is trigger via an external interrupt
on PD5 (PCINT21).

On start-up, the battery voltage is measured. If below a certain
threshold, the bi-colored LED blinks red on start-up, otherwise
it blinks green. Battery ADC voltage ratio is 0.848, meaning
max read voltage will be 2.544V.

Converter can work reliably down to 1.8V, or 0.9V per cell.
Alkaline batteries die at ca 0.8V, so to have a warning
when battery is at 15% capacity, trigger low voltage warning
when battery voltage is at 1.0V per cell, or 2.0V. With division
factor, this is 1.696V.

With reference of 3.3V, and 10-bit resolution, this is a 
reading less than 526.


Event detection pin:	PD4
External Buzzer:		PB3
Battery voltage ADC:	PC5
Battery gate enable:	PB0
Low voltage LED:		PC3
Voltage OK LED:			PC4
HC-05 key pin:			PC1
HC-05 prog en pin:		PC2


Project: GM Counter
MCU Type: ATmega328P
Clock: 7.3728MHz

Data Types:

int8_t - Signed Char 
uint16_t - Unsigned Int 
uint32_t - Unsigned Long 
int64_t - Signed Long Long

(signed/unsigned) char - 1 byte 
(signed/unsigned) short - 2 bytes 
(signed/unsigned) int - 2 bytes 
(signed/unsigned) long - 4 bytes 
(signed/unsigned) long long - 8 bytes 
float - 4 bytes (floating point) 
double - alias to float

******************************************************************/

// Battery voltage low definition
#define LOW_BAT_THRES 526
// Buzzer frequency setting compare match, sets the increment length. 1 = ca 34,72s
#define BUZZER_WATCHER_PERIOD 100
// How long a buzzer blip should last, in (1+n) * increment length
#define BLIP_DURATION 3
// Update interval compare match
#define UPDATE_PERIOD 0x707F
// LED positions
#define LED_PWR_LOW PC3
#define LED_PWR_GOOD PC4

// Flag definitions
#define READY 0
#define TX_ENABLED 1
#define RX_TRIGGERED 2

// Communication protocol definitions
// To GM counter
#define REQUEST_GM_TUBE_DATA "READC"
#define START_GM "START"
#define STOP_GM "HALTT"
// From MCU
#define HEADER_TIME "PERID:1000"
#define DATAGRAM_COUNT "COUNT:"

// SBM-19 specifics
/*#define HEADER_MAXCOUNT "MAXCT:2000"
#define HEADER_NAME "NAMET:SBM-19"
#define HEADER_DOSERATE "DOSER:875.0"*/

// SBM-20 specifics
#define HEADER_MAXCOUNT "MAXCT:2000"
#define HEADER_NAME "NAMET:SBM-20"
#define HEADER_DOSERATE "DOSER:175.0"

#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include "usart_lib.h"

// Declare global variables
volatile uint8_t tmr0_msb;             			// increments with every timer 0 overflow
volatile uint16_t counter;             			// Total number of events captured in 1 second
volatile uint8_t flag_reg;             			// Register for flags used in program
volatile int8_t buzzerTimeoutCntr;				// Used as a timeout for the buzzer

char stringBuffer[32];							// String buffer for copying out strings from prog mem

// Serial protocol constants
// Data sent from the MCU
const char DATAGRAM_STRING[] PROGMEM = DATAGRAM_COUNT;
const char NAME_STRING[] PROGMEM = HEADER_NAME;
const char TIME_STRING[] PROGMEM = HEADER_TIME;
const char MAXCOUNT_STRING[] PROGMEM = HEADER_MAXCOUNT;
const char DOSERATE_STRING[] PROGMEM = HEADER_DOSERATE;
// Data received by the MCU
/*const char REQUEST_CONFIG_STRING[] PROGMEM = REQUEST_GM_TUBE_DATA;
const char START_STRING[] PROGMEM = START_GM;
const char STOP_STRING[] PROGMEM = STOP_GM;*/

/*PGM_P const recv_protocol_string_table[] PROGMEM =
{
	REQUEST_CONFIG_STRING,
	START_STRING,
	STOP_STRING
};*/

PGM_P const send_protocol_string_table[] PROGMEM =
{
	NAME_STRING,
	TIME_STRING,
	MAXCOUNT_STRING,
	DOSERATE_STRING,
	DATAGRAM_STRING
};

// Setup the timers for counting events, and updating the main thread with new data
void initializeEventHardware(void);
// Setup timers and external interrupts for handling the buzzer
void initializeBuzzerHardwareItems(void);
// Trigger the buzzer
void triggerBuzzer(void);
// Set the device in a special mode for programming the HC-05 bluetooth module
void handleBluetoothProgrammingMode(void);
// Check the battery voltage at start-up
void checkBatteryVoltageLevel(void);
// Find Counts per minute
uint16_t count_math(uint16_t newCount);
// Write value to LCD
void displayNumberOnLCDLine(uint16_t data, uint8_t line);
// Parse and send a counts value
void sendCountValue(uint16_t count);
// Send the stored configuration over serial
void sendMyConfigurationData(void);
// Parse a character command
void parseCommandChar(char command);

// USART Received Data Complete interrupt service routine
ISR(USART_RX_vect) {
	// Place the USART contents in the buffer, so it can be handled later
	bufferInsert(UDR0);
}

// Timer 0 overflow interrupt service routine, used as a counter extension for detected events
ISR(TIMER0_OVF_vect)
{
    tmr0_msb++;
}

// Timer 1 Compare Match B interrupt service routine, used to keep track of how long the buzzer has been on
ISR(TIMER1_COMPA_vect)
{
	// Disable Interrupts
	cli();
	
	// Run the timer state machine
	if (buzzerTimeoutCntr > BLIP_DURATION)
	{
		// Disable buzzer
		TCCR2B &= ~((1 << CS22)|(1 << CS21)|(1 << CS20));
		PORTB &= ~(1 << PB3);
		// Pause the counter
		buzzerTimeoutCntr = -1;
	}
	else if (buzzerTimeoutCntr >= 0)
	{
		buzzerTimeoutCntr++;
	}
	
	// Move the next compare match time into the future
	OCR1A = TCNT1 + BUZZER_WATCHER_PERIOD;
	sei();
}

// Timer 1 Compare Match B interrupt service routine, used to synchronize serial communication
ISR(TIMER1_COMPB_vect)
{
	// Disable Interrupts
	cli();
	
	// This routine is entered at the update interval of the device, right now hard coded to 1000ms
    
	// Find total number of events during last second
    counter = (255 * tmr0_msb) + TCNT0;
	
	// Reinitialize Timer values
	TCNT0 = 0;
	tmr0_msb = 0;
	
	// Move the next compare match time into the future. Notify main thread that new data is ready
	flag_reg |= (1 << READY);
    OCR1B = TCNT1 + UPDATE_PERIOD;
	
	// Enable Interrupts
	sei();
}

// Used to trigger the buzzer, on rising edges
ISR(INT1_vect)
{
	// Disable Interrupts
	cli();
	
	triggerBuzzer();
	
	sei();
}

// Main
int main(void)
{
	char latestReceivedByte;

	// Turn off analog comparator to save power
	ACSR = (1<<ACD);
	
	// Set PORTB output direction for power-up voltage measurement (PB0) and buzzer control (PB3)
	DDRB = (1 << PB3)|(1 << PB0);
	// Set high to keep the battery voltage gated off, and buzzer low to keep disabled
	PORTB = (1 << PB0);

	// Set PORTD and PORTC pull-ups. Set to a one to enable
	PORTD = 0;
	PORTC = (1 << PC2);
	
	checkBatteryVoltageLevel();
	
	// Check if HC-05 programming mode is enabled. If so, simply set the KEY pin high, and do nothing.
	handleBluetoothProgrammingMode();
	
	// Initialize the hardware modules needed otherwise
	initializeEventHardware();
	initializeBuzzerHardwareItems();
	USART_Init();
	
	// Set default mode
	flag_reg &= ~(1 << TX_ENABLED);
	flag_reg |= (1 << RX_TRIGGERED);	// Ready to act on the first byte received

	sei();
	
	triggerBuzzer();
	
	while(1)
	{
		// New byte received
		if (writeIndex != currentIndex)
		{
			latestReceivedByte = popFromBuffer();
			
			// Once a LF char is received, the next character should be the start of a new command
			// No other data is needed. Check if the last char was a LF, if so use the data.
			// Otherwise discard it.
			if (bit_is_set(flag_reg, RX_TRIGGERED))
			{
				// Act on the command received
				parseCommandChar(latestReceivedByte);
				flag_reg &= ~(1 << RX_TRIGGERED);
			}
			else if (latestReceivedByte == '\n')
			{
				// Ready to act on the next byte received
				flag_reg |= (1 << RX_TRIGGERED);
			}
			
		}
		
		
		// Simple check so display in only updated once per second
		if (bit_is_set(flag_reg, READY))
		{
			// New count data, display and send out via UART
			if (bit_is_set(flag_reg, TX_ENABLED))
			{
				sendCountValue(counter);
			}
			
			// Clear ready flag
			flag_reg  &=~(1 << READY);
		}
		
	}
	
	return 0;
}

// Setup the timers for counting events, and updating the main thread with new data
void initializeEventHardware(void)
{
	// Timer/Counter 0 initialization, set clock source to T0 pin rising Edge
	TCCR0B = (1<<CS02)|(1<<CS01)|(1<<CS00);
	counter = 0;
	
	// Timer/Counter 1 initialization
	// Timer is used both to set update interval, and to control the buzzer tone
	// Set Prescaler to 1:256, and compare match B every 28800 cycles, for 1000ms interrupt rate. This means timer compare must happen at 28799 (0x707F)
	OCR1B = UPDATE_PERIOD;
	
	// Set the clock source
	TCCR1B = (1<<CS12)|(0<<CS11)|(0<<CS10);
	
	// Timer 0 and 1 Interrupt initialization
	TIMSK1 = (1<<OCIE1B);
	TIMSK0 = (1<<TOIE0);
}

// Setup timers and external interrupts for handling the buzzer
void initializeBuzzerHardwareItems(void)
{
	// Synchronous mode, clocked from main system clock
	ASSR &= ~(1 << AS2);
	// Toggle OC2A on compare match, CTC mode
	TCCR2A = (0 << COM2A1)|(1 << COM2A0)|(1 << WGM21)|(0 << WGM20);
	TCCR2B = (0 << WGM22);
	
	// Compare match A on Timer1 is set as the watch dog, so Timer2 is only active for a short duration
	OCR1A = BUZZER_WATCHER_PERIOD;
	TIMSK1 |= (1<<OCIE1A);
	
	// Setup external interrupt capturing on INT1 rising edge
	EICRA = (1 << ISC11)|(1 << ISC10);
	EIMSK = (1 << INT1);
	
	// Set to paused state
	buzzerTimeoutCntr = -1;
}

// Trigger the buzzer
void triggerBuzzer(void)
{
	// Enable buzzer, Set to a frequency near 4186.01Hz (high C, and also near center for the piezo element), using a prescaler of 1:8 and compare of 220
	TCCR2B |= (0 << CS22)|(1 << CS21)|(0 << CS20);
	OCR2A = 220;
	
	// Start the buzzer state machine
	buzzerTimeoutCntr = 0;
}

// Set the device in a special mode for programming the HC-05 bluetooth module
void handleBluetoothProgrammingMode(void)
{
	if (!(PINC & (1 << PC2)))
	{
		// Set the KEY pin high
		DDRC = (1 << PC1);
		PORTC = (1 << PC1);
		
		// Wait until reset
		// Alternate between blinking the RED and GREEN LEDs, then sounding buzzer briefly, doubles as a hardware test
		DDRC = (1 << LED_PWR_GOOD)|(1 << LED_PWR_LOW);
		
		while(1)
		{
			PORTC &= ~(1 << LED_PWR_GOOD);
			PORTC |= (1 << LED_PWR_LOW);
			_delay_ms(500);
			
			PORTC &= ~(1 << LED_PWR_LOW);
			PORTC |= (1 << LED_PWR_GOOD);
			_delay_ms(500);
			
			triggerBuzzer();
		}
	}
}

// Check the battery voltage at start-up
void checkBatteryVoltageLevel(void)
{
	#define AVERAGING_PASSES 5
	
	uint16_t averageStorage = 0;
	uint16_t batteryVoltageReading = 0;
	uint8_t i;
	
	// Initialize ADC, AVCC (3.3V) as reference, and set ADC input to PC5
	ADMUX = (0 << REFS1)|(1 << REFS0)|(PC5);
	
	// Enable ADC, and set prescaler to 1:32 for 230.4kHz operation.
	ADCSRA = (1 << ADEN)|(1 << ADPS2)|(0 << ADPS1)|(1 << ADPS0);
	
	// Gate on the battery voltage, and wait for it to stabilize
	PORTB &= ~(1 << PB0);
	_delay_ms(50);
	
	// Average several readings to get a low noise figure
	for(i=0; i<AVERAGING_PASSES; i++)
	{
		// Start a conversion
		ADCSRA |= (1 << ADSC);
		// Wait for conversion to finish
		while(!(ADCSRA & (1 << ADIF)));
		// Clear ADC complete flag
		ADCSRA |= (1 << ADIF);
		// Fetch the ADC result
		averageStorage += ADC;
	}
	
	// Reading from ADC complete, so disable it
	ADCSRA &= ~(1 << ADEN);
	
	// Turn off the battery voltage gating
	PORTB = (1 << PB0);
	
	// Find the average
	batteryVoltageReading = averageStorage / AVERAGING_PASSES;
	
	// Parse ADC results
	if (batteryVoltageReading < LOW_BAT_THRES)
	{
		// Battery voltage is low, blink the RED LED
		DDRC = (1 << LED_PWR_LOW);
		PORTC = (1 << LED_PWR_LOW);
	}
	else
	{
		// Battery voltage is OK, blink the GREEN LED
		DDRC = (1 << LED_PWR_GOOD);
		PORTC = (1 << LED_PWR_GOOD);
	}
	_delay_ms(1000);
	
	// Clear the LEDs
	DDRC &= ~((1 << LED_PWR_GOOD)|(1 << LED_PWR_LOW));
	PORTC &= ~((1 << LED_PWR_GOOD)|(1 << LED_PWR_LOW));
}

// Parse and send a counts value
void sendCountValue(uint16_t count)
{
	char countString[7];	// uint16_t, so at most 5 characters, plus one termination char. Add one extra for good measure.
	
	// Place the datagram header in the string
	strcpy_P(stringBuffer, (PGM_P)pgm_read_word(&(send_protocol_string_table[4])));

	// Format output
	utoa(count, countString, 10);
	strcat(stringBuffer, countString);
	
	// Send the string
	sendSerialString(stringBuffer);
}

// Send the stored configuration over serial
void sendMyConfigurationData(void)
{
	// Get the correct strings from prog mem and send them
	for (unsigned char i = 0; i < 4; i++)
	{
		strcpy_P(stringBuffer, (PGM_P)pgm_read_word(&(send_protocol_string_table[i])));
		sendSerialString(stringBuffer);
	}
}

// Parse a character command
void parseCommandChar(char command)
{
	// Only need to check the first letter of the command to know what it is
	switch (command)
	{
		case 'R':
			// Request for configuration received
			sendMyConfigurationData();
			break;
			
		case 'S':
			// Start transmitting command received
			flag_reg |= (1 << TX_ENABLED);
			break;
			
		case 'H':
			// Stop transmitting command received
			flag_reg &= ~(1 << TX_ENABLED);
			break;
	}
}
